@@ -5,7 +5,7 @@
 ;; Author: Protesilaos Stavrou <info@protesilaos.com>
 ;; Maintainer: Protesilaos Stavrou <info@protesilaos.com>
 ;; URL: https://github.com/protesilaos/beframe
-;; Version: 1.3.0
+;; Version: 1.4.0
 ;; Package-Requires: ((emacs "28.1"))
 
 ;; This file is NOT part of GNU Emacs.
@@ -66,6 +66,16 @@ buffers from a frame's buffer list ad-hoc.  The full list of commands:
                    (symbol :tag "Symbol to match a buffer's major mode" :value "")))
           (const :tag "No global buffers" nil)))
 
+(defcustom beframe-prompt-prefix "[Beframed]"
+  "String prefix for buffer prompts affected by `beframe-mode'.
+Set to nil or an empty string to have no prefix.
+
+The text is styled with the face `beframe-face-prompt-prefix'."
+  :group 'beframe
+  :package-version '(beframe . "1.4.0")
+  :type '(choice (string :tag "Text prefix for buffer prompts")
+                 (const :tag "No prefix" nil)))
+
 (defcustom beframe-create-frame-scratch-buffer t
   "Create a frame-specific scratch buffer for new frames.
 Do it when `beframe-mode' is enabled.
@@ -140,6 +150,16 @@ has a running process, and the like."
   :package-version '(beframe . "1.2.0")
   :type 'boolean)
 
+(defgroup beframe-faces nil
+  "Faces used by Beframe."
+  :group 'beframe
+  :group 'faces)
+
+(defface beframe-face-prompt-prefix '((t :inherit italic))
+  "Face applied to `beframe-prompt-prefix'."
+  :group 'beframe-faces
+  :package-version '(beframe . "1.4.0"))
+
 (defun beframe--remove-internal-buffers (buffers)
   "Remove internal buffers from BUFFERS."
   (seq-remove
@@ -166,6 +186,31 @@ through `beframe--remove-internal-buffers'."
          (string-match-p regexp (buffer-name buffer))))
      buffers)))
 
+(defun beframe--parse-global-buffers ()
+  "Parse `beframe-global-buffers' into (regexp . symbols).
+Make strings a single regular expression out of all strings and collect
+any symbols into a list."
+  (let ((strings nil)
+        (symbols nil))
+    (dolist (element beframe-global-buffers)
+      (or (and (stringp element) (push element strings))
+          (and (symbolp element) (push element symbols))))
+    (cons (string-join strings "\\|") symbols)))
+
+(defun beframe--global-buffers ()
+  "Return list of `beframe-global-buffers' buffer objects."
+  (pcase-let* ((public-buffers (beframe--get-buffers 'public))
+               (global-buffers nil)
+               (`(,regexp . ,modes) (beframe--parse-global-buffers)))
+    (dolist (buffer public-buffers)
+      (cond
+       ((string-match-p regexp (buffer-name buffer))
+        (push buffer global-buffers))
+       ((with-current-buffer buffer
+          (when (derived-mode-p modes)
+            (push buffer global-buffers))))))
+    global-buffers))
+
 (defun beframe--get-buffers (&optional arg)
   "Return list of buffers from different sources depending on ARG.
 
@@ -200,19 +245,6 @@ The following values of ARG can be used:
      (beframe--remove-internal-buffers (frame-parameter arg 'buffer-list)))
     (_ (user-error "Wrong argument in `beframe--get-buffers' pcase"))))
 
-(defun beframe--global-buffers ()
-  "Return list of `beframe-global-buffers' buffer objects."
-  (pcase-let* ((pub-buffs (beframe--get-buffers 'public))
-               (`(,str-re ,modes-lst)
-                (cl-loop for E in beframe-global-buffers
-                         if (stringp E) collect E into str
-                         else if (symbolp E) collect E into sym
-                         finally return (list (string-join str "\\|") sym))))
-    (cl-loop for b in pub-buffs
-             if (string-match-p str-re (buffer-name b)) collect b
-             else if (with-current-buffer b (derived-mode-p modes-lst))
-             collect b)))
-
 (cl-defun beframe-buffer-list (&optional frame &key sort)
   "Return list of buffers that are used by the current frame.
 With optional FRAME as an object that satisfies `framep', return
@@ -251,33 +283,42 @@ more information."
   'beframe-buffer-names
   "0.2.0")
 
-(defun beframe--read-buffer-p (buf &optional frame)
-  "Return non-nil if BUF belongs to the current FRAME.
-BUF is a string or a cons cell, per `beframe-read-buffer'.
-If optional FRAME is nil, then default to the current one.  Else
-it must satisfy `framep'."
-  (let ((b buf))
-    (when (consp buf)
-      (setq b (car buf)))
-    (unless (string-prefix-p " " b)
-      (seq-contains-p (beframe-buffer-names frame) b))))
+(defun beframe--read-buffer-p (buffer &optional frame)
+  "Return non-nil if BUFFER belongs to the current FRAME.
+BUFFER is a string or a cons cell, per `beframe-read-buffer'.  If
+optional FRAME is nil, then default to the current one.  Else FRAME is
+an object that satisfies `framep'."
+  (when (consp buffer)
+    (setq buffer (car buffer)))
+  (unless (string-prefix-p " " buffer)
+    (seq-contains-p (beframe-buffer-names frame) buffer)))
 
 (defvar beframe-history nil
   "Minibuffer history of frame specific buffers.")
 
+(defun beframe--propertize-prompt-prefix ()
+  "Return `beframe-prompt-prefix' with face `beframe-face-prompt-prefix'.
+If `beframe-prompt-prefix' is nil or an empty string, then return an
+empty string."
+  (if (and (stringp beframe-prompt-prefix)
+           (not (string-blank-p beframe-prompt-prefix)))
+      (format "​%s " (propertize beframe-prompt-prefix 'face 'beframe-face-prompt-prefix))
+    ""))
+
 ;;;###autoload
 (defun beframe-read-buffer (prompt &optional def require-match _predicate)
   "The `read-buffer-function' that limits buffers to frames.
 PROMPT, DEF, REQUIRE-MATCH, and PREDICATE are the same as
 `read-buffer'.  The PREDICATE is ignored, however, to apply the
 per-frame filter."
-  (completing-read (format "[Beframed] %s" prompt)
-                   (beframe-buffer-names)
-                   #'beframe--read-buffer-p
-                   require-match
-                   nil
-                   'beframe-history
-                   def))
+  (completing-read
+   (format "%s%s" (beframe--propertize-prompt-prefix) prompt)
+   (beframe-buffer-names)
+   #'beframe--read-buffer-p
+   require-match
+   nil
+   'beframe-history
+   def))
 
 (defun beframe--buffer-prompt (&optional frame)
   "Prompt for buffer among `beframe-buffer-names'.
@@ -729,6 +770,24 @@ Meant to be assigned to a prefix key, like this:
 (define-key beframe-prefix-map (kbd "a s") #'beframe-assume-buffers-matching-regexp)
 (define-key beframe-prefix-map (kbd "u s") #'beframe-unassume-buffers-matching-regexp)
 
+(defvar xref-history-storage)
+
+(defvar beframe-xref-history-storage (bound-and-true-p xref-history-storage)
+  "Store the last known value of `xref-history-storage'.")
+
+(defun beframe-xref-frame-history (&optional new-value)
+  "Return frame-specific Xref history for the current frame.
+Override existing value with NEW-VALUE if NEW-VALUE is set.
+
+This function is based on `xref-window-local-history'."
+  (let ((frame (selected-frame)))
+    (cond
+     (new-value
+      (set-frame-parameter frame 'xref--history new-value))
+     ((frame-parameter frame 'xref--history))
+     (t
+      (beframe-create-xref-history frame)))))
+
 ;;;###autoload
 (define-minor-mode beframe-mode
   "Make all buffer prompts limit candidates per frame.
@@ -737,12 +796,14 @@ Also see the variable `beframe-prefix-map'."
   (if beframe-mode
       (progn
         (setq beframe--read-buffer-function read-buffer-function
-              read-buffer-function #'beframe-read-buffer)
+              read-buffer-function #'beframe-read-buffer
+              xref-history-storage #'beframe-xref-frame-history)
         (add-hook 'after-make-frame-functions #'beframe-setup-frame)
         (add-hook 'context-menu-functions #'beframe-context-menu)
         (beframe--functions-in-frames))
     (setq read-buffer-function beframe--read-buffer-function
-          beframe--read-buffer-function nil)
+          beframe--read-buffer-function nil
+          xref-history-storage beframe-xref-history-storage)
     (remove-hook 'after-make-frame-functions #'beframe-setup-frame)
     (remove-hook 'context-menu-functions #'beframe-context-menu)
     (beframe--functions-in-frames :disable)))
@@ -893,6 +954,13 @@ FRAME and optional NAME arguments are passed to the
   (when beframe-rename-function
     (funcall beframe-rename-function frame name)))
 
+(defun beframe-create-xref-history (frame)
+  "Create new Xref history for FRAME unless it already exists."
+  (unless (frame-parameter frame 'xref--history)
+    (let ((history (cons nil nil)))
+      (set-frame-parameter frame 'xref--history history)
+      history)))
+
 (defun beframe-setup-frame (frame)
   "Rename FRAME and create scratch buffer for it, if appropriate.
 Call the functions `beframe-frame-predicate',
@@ -902,7 +970,8 @@ this order."
   (dolist (fn '(beframe-frame-predicate
                 beframe-do-not-assume-last-selected-buffer
                 beframe-maybe-rename-frame
-                beframe-create-scratch-buffer))
+                beframe-create-scratch-buffer
+                beframe-create-xref-history))
     (funcall fn frame)))
 
 (defun beframe--frame-buffer-p (buf &optional frame)
@@ -1002,7 +1071,6 @@ This function can be used as the :sort key of
   "Menu with all Beframe commands, each available in the right context."
   beframe--menu-contents)
 
-;; Add Beframe menu at the end of global-map after Tools
 (easy-menu-add-item global-map '(menu-bar "Buffers") beframe-global-menu)
 
 (defun beframe-context-menu (menu _click)
